import os
import random
import sys
import time
import unittest
from imp import reload

import z3c.pt
import z3c.ptcompat.engine
import zope.component
import zope.component.globalregistry
import zope.configuration.xmlconfig
import zope.interface
import zope.schema
from z3c.pt.pagetemplate import ViewPageTemplateFile as z3cViewPageTemplateFile
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.pagetemplate.interfaces import IPageTemplateEngine

from z3c.form import field
from z3c.form import form
from z3c.form import term
from z3c.form import testing
from z3c.form import tests


reload(sys)
sys.setdefaultencoding('utf-8')


def benchmark(title):
    def decorator(f):
        def wrapper(*args):
            print("===============================")
            print(title)
            print("===============================")
            return f(*args)
        return wrapper
    return decorator


def timing(func, *args, **kwargs):
    t1 = t2 = time.time()
    i = 0
    while t2 - t1 < 3:
        func(*args, **kwargs)
        i += 1
        t2 = time.time()
    return 100 * (t2 - t1) / i


class ISmallForm(zope.interface.Interface):
    name = zope.schema.TextLine(
        title="Name",
        description="Please enter your first and last name.")

    address = zope.schema.TextLine(
        title="Address",
        description="Please enter a valid address.")


class ILargeDataSetsForm(zope.interface.Interface):
    lucky_numer = zope.schema.Choice(
        range(500),
        title="Lucky number",
        description="Choose your lucky number.")

    favorite_letters = zope.schema.Set(
        title="Favorite letter",
        description="Choose your favorite letter.",
        value_type=zope.schema.Choice(
            ["".join(chr(random.randint(65, 90)) for i in range(10))]
        ))


def build_many_fields(size):
    for i in range(size):
        name = "".join(chr(random.randint(65, 90)) for i in range(10))
        yield zope.schema.TextLine(
            __name__=name,
            description="This field renders %s" % name,
            title="Title of %s" % name.capitalize())


IManyFields = tuple(build_many_fields(500))


class BaseTestCase(unittest.TestCase):
    def _setUp(suite):
        testing.setUp(suite)
        testing.setupFormDefaults()
        zope.configuration.xmlconfig.XMLConfig('configure.zcml', z3c.pt)()
        zope.component.provideAdapter(term.CollectionTerms)

    def _tearDown(suite):
        testing.tearDown(suite)


def enableZ3CPT():
    """Enable z3c.pt engine"""
    base = zope.component.globalregistry.base
    base.registerUtility(z3c.ptcompat.engine.Program, IPageTemplateEngine,
                         name='', event=False)


def disableZ3CPT():
    """Disable z3c.pt engine"""
    base = zope.component.globalregistry.base
    base.unregisterUtility(z3c.ptcompat.engine.Program, IPageTemplateEngine,
                           name='')


class BenchmarkTestCase(BaseTestCase):
    def simple_form(self, cls, iface):
        class SimpleForm(form.AddForm):
            fields = field.Fields(
                *(isinstance(iface, tuple) and iface or (iface,)))
            render = cls(
                os.path.join(tests.__path__[0], 'simple_edit.pt'))
        return SimpleForm

    @benchmark("Small add-form (update/render)")
    def testSmallForm(self):
        context = object()
        request = testing.TestRequest()

        f_z3c = self.simple_form(
            z3cViewPageTemplateFile, ISmallForm)(context, request)
        f_zope = self.simple_form(
            ViewPageTemplateFile, ISmallForm)(context, request)

        t_z3c = self.benchmark(enableZ3CPT, f_z3c)
        t_zope = self.benchmark(disableZ3CPT, f_zope)

        print("z3c.pt:            %.3f" % t_z3c)
        print("zope.pagetemplate: %.3f" % t_zope)
        print("                   %.2fX" % (t_zope / t_z3c))

    @benchmark("Large data sets (update/render)")
    def testLargeDataSets(self):
        context = object()
        request = testing.TestRequest()

        f_z3c = self.simple_form(
            z3cViewPageTemplateFile, ILargeDataSetsForm)(context, request)
        f_zope = self.simple_form(
            ViewPageTemplateFile, ILargeDataSetsForm)(context, request)

        t_z3c = self.benchmark(enableZ3CPT, f_z3c)
        t_zope = self.benchmark(disableZ3CPT, f_zope)

        print("z3c.pt:            %.3f" % t_z3c)
        print("zope.pagetemplate: %.3f" % t_zope)
        print("                   %.2fX" % (t_zope / t_z3c))

    @benchmark("Many fields (update/render)")
    def testManyFields(self):
        context = object()
        request = testing.TestRequest()

        f_z3c = self.simple_form(
            z3cViewPageTemplateFile, IManyFields)(context, request)
        f_zope = self.simple_form(
            ViewPageTemplateFile, IManyFields)(context, request)

        t_z3c = self.benchmark(enableZ3CPT, f_z3c)
        t_zope = self.benchmark(disableZ3CPT, f_zope)

        print("z3c.pt:            %.3f" % t_z3c)
        print("zope.pagetemplate: %.3f" % t_zope)
        print("                   %.2fX" % (t_zope / t_z3c))

    def benchmark(self, prep, func, *args):
        self._setUp()
        func(*args)
        prep()
        func(*args)
        t = timing(func, *args)
        self._tearDown()
        return t


def test_suite():
    return unittest.TestSuite((
        unittest.defaultTestLoader.loadTestsFromTestCase(BenchmarkTestCase),
    ))


if __name__ == "__main__":
    unittest.main(defaultTest="test_suite")



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Error Views Implementation

$Id$
"""
__docformat__ = "reStructuredText"
import os

import zope.component
import zope.interface
import zope.schema
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.pagetemplate.interfaces import IPageTemplate

import z3c.form
from z3c.form import interfaces
from z3c.form import util
from z3c.form import value
from z3c.form.i18n import MessageFactory as _


ErrorViewMessage = value.StaticValueCreator(
    discriminators=('error', 'request', 'widget', 'field', 'form', 'content')
)

ComputedErrorViewMessage = value.ComputedValueCreator(
    discriminators=('error', 'request', 'widget', 'field', 'form', 'content')
)


def ErrorViewDiscriminators(
        errorView,
        error=None,
        request=None,
        widget=None,
        field=None,
        form=None,
        content=None):
    zope.component.adapter(
        util.getSpecification(error),
        util.getSpecification(request),
        util.getSpecification(widget),
        util.getSpecification(field),
        util.getSpecification(form),
        util.getSpecification(content))(errorView)


@zope.interface.implementer(interfaces.IErrorViewSnippet)
class ErrorViewSnippet:
    """Error view snippet."""
    zope.component.adapts(
        zope.schema.ValidationError, None, None, None, None, None)

    def __init__(self, error, request, widget, field, form, content):
        self.error = self.context = error
        self.request = request
        self.widget = widget
        self.field = field
        self.form = form
        self.content = content

    def createMessage(self):
        return self.error.doc()

    def update(self):
        value = zope.component.queryMultiAdapter(
            (self.context, self.request, self.widget,
             self.field, self.form, self.content),
            interfaces.IValue, name='message')
        if value is not None:
            self.message = value.get()
        else:
            self.message = self.createMessage()

    def render(self):
        template = zope.component.getMultiAdapter(
            (self, self.request), IPageTemplate)
        return template(self)

    def __repr__(self):
        return '<{} for {}>'.format(
            self.__class__.__name__, self.error.__class__.__name__)


class ValueErrorViewSnippet(ErrorViewSnippet):
    """An error view for ValueError."""
    zope.component.adapts(
        ValueError, None, None, None, None, None)

    defaultMessage = _('The system could not process the given value.')

    def createMessage(self):
        return self.defaultMessage


class InvalidErrorViewSnippet(ErrorViewSnippet):
    """Error view snippet."""
    zope.component.adapts(
        zope.interface.Invalid, None, None, None, None, None)

    def createMessage(self):
        return self.error.args[0]


class MultipleErrorViewSnippet(ErrorViewSnippet):
    """Error view snippet for multiple errors."""
    zope.component.adapts(
        interfaces.IMultipleErrors, None, None, None, None, None)

    def update(self):
        pass

    def render(self):
        return ''.join([view.render() for view in self.error.errors])


@zope.interface.implementer(interfaces.IMultipleErrors)
class MultipleErrors(Exception):
    """An error that contains many errors"""

    def __init__(self, errors):
        self.errors = errors


class ErrorViewTemplateFactory:
    """Error view template factory."""

    template = None

    def __init__(self, filename, contentType='text/html'):
        self.template = ViewPageTemplateFile(
            filename, content_type=contentType)

    def __call__(self, errorView, request):
        return self.template


# Create the standard error view template
StandardErrorViewTemplate = ErrorViewTemplateFactory(
    os.path.join(os.path.dirname(z3c.form.__file__), 'error.pt'), 'text/html')
zope.component.adapter(
    interfaces.IErrorViewSnippet, None)(StandardErrorViewTemplate)
zope.interface.implementer(IPageTemplate)(StandardErrorViewTemplate)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""ObjectWidget related classes

$Id$
"""
__docformat__ = "reStructuredText"
import zope.component
import zope.event
import zope.interface
import zope.lifecycleevent
import zope.schema
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.pagetemplate.interfaces import IPageTemplate
from zope.security.proxy import removeSecurityProxy

from z3c.form import field
from z3c.form import interfaces
from z3c.form import util
from z3c.form import widget
from z3c.form.converter import BaseDataConverter
from z3c.form.error import MultipleErrors


def getIfName(iface):
    return iface.__module__ + '.' + iface.__name__


# our own placeholder instead of a simple None
class ObjectWidget_NO_VALUE:
    def __repr__(self):
        return '<ObjectWidget_NO_VALUE>'


ObjectWidget_NO_VALUE = ObjectWidget_NO_VALUE()


class ObjectWidgetValue(dict):
    originalValue = ObjectWidget_NO_VALUE  # will store the original object


class ObjectConverter(BaseDataConverter):
    """Data converter for IObjectWidget."""

    zope.component.adapts(
        zope.schema.interfaces.IObject, interfaces.IObjectWidget)

    def toWidgetValue(self, value):
        """Just dispatch it."""
        if value is self.field.missing_value:
            return interfaces.NO_VALUE

        retval = ObjectWidgetValue()
        retval.originalValue = value

        for name, field_ in zope.schema.getFieldsInOrder(self.field.schema):
            dm = zope.component.getMultiAdapter(
                (value, field_), interfaces.IDataManager)
            subv = dm.query()

            if subv is interfaces.NO_VALUE:
                # look up default value
                subv = field_.default
                # XXX: too many discriminators
                # adapter = zope.component.queryMultiAdapter(
                #    (context, self.request, self.view, field, widget),
                #    interfaces.IValue, name='default')
                # if adapter:
                #    value = adapter.get()

            widget = zope.component.getMultiAdapter(
                (field_, self.widget.request), interfaces.IFieldWidget)
            if interfaces.IFormAware.providedBy(self.widget):
                # form property required by objectwidget
                widget.form = self.widget.form
                zope.interface.alsoProvides(widget, interfaces.IFormAware)
            converter = zope.component.getMultiAdapter(
                (field_, widget), interfaces.IDataConverter)

            retval[name] = converter.toWidgetValue(subv)

        return retval

    def adapted_obj(self, obj):
        return self.field.schema(obj)

    def toFieldValue(self, value):
        """field value is an Object type, that provides field.schema"""
        if value is interfaces.NO_VALUE:
            return self.field.missing_value

        # try to get the original object, or if there's no chance an empty one
        obj = self.widget.getObject(value)
        obj = self.adapted_obj(obj)

        names = []
        for name, field_ in zope.schema.getFieldsInOrder(self.field.schema):
            if not field_.readonly:
                try:
                    newvalRaw = value[name]
                except KeyError:
                    continue

                widget = zope.component.getMultiAdapter(
                    (field_, self.widget.request), interfaces.IFieldWidget)
                converter = zope.component.getMultiAdapter(
                    (field_, widget), interfaces.IDataConverter)
                newval = converter.toFieldValue(newvalRaw)

                dm = zope.component.getMultiAdapter(
                    (obj, field_), interfaces.IDataManager)
                oldval = dm.query()
                if (oldval != newval
                        or zope.schema.interfaces.IObject.providedBy(field_)):
                    dm.set(newval)
                    names.append(name)

        if names:
            zope.event.notify(
                zope.lifecycleevent.ObjectModifiedEvent(
                    obj, zope.lifecycleevent.Attributes(
                        self.field.schema, *names)))

        # Commonly the widget context is security proxied. This method,
        # however, should return a bare object, so let's remove the
        # security proxy now that all fields have been set using the security
        # mechanism.
        return removeSecurityProxy(obj)


@zope.interface.implementer(interfaces.IObjectWidget)
class ObjectWidget(widget.Widget):

    _mode = interfaces.INPUT_MODE
    _value = interfaces.NO_VALUE
    _updating = False
    prefix = ''
    widgets = None

    def createObject(self, value):
        # keep value passed, maybe some subclasses want it
        # value here is the raw extracted from the widget's subform
        # in the form of a dict key:fieldname, value:fieldvalue
        name = getIfName(self.field.schema)
        creator = zope.component.queryMultiAdapter(
            (self.context, self.request, self.form, self),
            interfaces.IObjectFactory,
            name=name)
        if creator:
            obj = creator(value)
        else:
            # raise RuntimeError, that won't be swallowed
            raise RuntimeError(
                "No IObjectFactory adapter registered for %s" % name)

        return obj

    def getObject(self, value):
        if value.originalValue is ObjectWidget_NO_VALUE:
            # if the originalValue did not survive the roundtrip
            if self.ignoreContext:
                obj = self.createObject(value)
            else:
                # try to get the original object from the context.field_name
                dm = zope.component.getMultiAdapter(
                    (self.context, self.field), interfaces.IDataManager)
                try:
                    obj = dm.get()
                except KeyError:
                    obj = self.createObject(value)
                except AttributeError:
                    obj = self.createObject(value)
        else:
            # reuse the object that we got in toWidgetValue
            obj = value.originalValue

        if obj is None or obj == self.field.missing_value:
            # if still None, create one, otherwise following will burp
            obj = self.createObject(value)

        return obj

    @property
    def mode(self):
        """This sets the subwidgets modes."""
        return self._mode

    @mode.setter
    def mode(self, mode):
        self._mode = mode
        # ensure that we apply the new mode to the widgets
        if self.widgets:
            for w in self.widgets.values():
                w.mode = mode

    def setupFields(self):
        self.fields = field.Fields(self.field.schema)

    def setupWidgets(self):
        self.setupFields()

        self.prefix = self.name
        self.widgets = field.FieldWidgets(self, self.request, None)
        self.widgets.mode = self.mode
        # very-very important! otherwise the update() tries to set
        # RAW values as field values
        self.widgets.ignoreContext = True
        self.widgets.ignoreRequest = self.ignoreRequest
        self.widgets.update()

    def updateWidgets(self, setErrors=True):
        if self.field is None:
            raise ValueError(
                "%r .field is None, that's a blocking point" %
                self)

        self.setupWidgets()

        if self._value is interfaces.NO_VALUE:
            # XXX: maybe readonly fields/widgets should be reset here to
            #      widget.mode = INPUT_MODE
            pass
            for name, widget_ in self.widgets.items():
                if widget_.field.readonly:
                    widget_.mode = interfaces.INPUT_MODE
                    widget_.update()
        else:
            rawvalue = None

            for name, widget_ in self.widgets.items():
                if widget_.mode == interfaces.DISPLAY_MODE:
                    if rawvalue is None:
                        # lazy evaluation
                        converter = zope.component.getMultiAdapter(
                            (self.field, self),
                            interfaces.IDataConverter)
                        obj = self.getObject(self._value)
                        rawvalue = converter.toWidgetValue(obj)

                    self.applyValue(widget_, rawvalue[name])
                else:
                    try:
                        v = self._value[name]
                    except KeyError:
                        pass
                    else:
                        self.applyValue(widget_, v)

    def applyValue(self, widget, value):
        """Validate and apply value to given widget.

        This method gets called on any ObjectWidget value change and is
        responsible for validating the given value and setup an error message.

        This is internal apply value and validation process is needed because
        nothing outside this widget does know something about our
        internal sub widgets.
        """
        if value is not interfaces.NO_VALUE:
            try:
                # convert widget value to field value
                converter = interfaces.IDataConverter(widget)
                fvalue = converter.toFieldValue(value)
                # validate field value
                zope.component.getMultiAdapter(
                    (self.context,
                     self.request,
                     self.form,
                     getattr(widget, 'field', None),
                     widget),
                    interfaces.IValidator).validate(fvalue)
                # convert field value back to widget value
                # that will probably format the value too
                widget.value = converter.toWidgetValue(fvalue)
            except (zope.schema.ValidationError, ValueError) as error:
                # on exception, setup the widget error message
                view = zope.component.getMultiAdapter(
                    (error, self.request, widget, widget.field,
                     self.form, self.context), interfaces.IErrorViewSnippet)
                view.update()
                widget.error = view
                # set the wrong value as value despite it's wrong
                # we want to re-show wrong values
                widget.value = value

    def update(self):
        # very-very-nasty: skip raising exceptions in extract while we're
        # updating
        self._updating = True
        try:
            super().update()
            # create the subwidgets and set their values
            self.updateWidgets(setErrors=False)
        finally:
            self._updating = False

    @property
    def value(self):
        # value (get) cannot raise an exception, then we return insane values
        try:
            self.setErrors = True
            return self.extract()
        except MultipleErrors:
            value = ObjectWidgetValue()
            if self._value is not interfaces.NO_VALUE:
                # send back the original object
                value.originalValue = self._value.originalValue

            for name, widget_ in self.widgets.items():
                if widget_.mode != interfaces.DISPLAY_MODE:
                    value[name] = widget_.value
            return value

    @value.setter
    def value(self, value):
        # This invokes updateWidgets on any value change e.g. update/extract.
        if (not isinstance(value, ObjectWidgetValue)
                and value is not interfaces.NO_VALUE):
            value = ObjectWidgetValue(value)
        self._value = value

        # create the subwidgets and set their values
        self.updateWidgets()

    def extractRaw(self, setErrors=True):
        '''See interfaces.IForm'''
        self.widgets.setErrors = setErrors
        # self.widgets.ignoreRequiredOnExtract = self.ignoreRequiredOnExtract
        data, errors = self.widgets.extractRaw()
        value = ObjectWidgetValue()
        if self._value is not interfaces.NO_VALUE:
            # send back the original object
            value.originalValue = self._value.originalValue
        value.update(data)

        return value, errors

    def extract(self, default=interfaces.NO_VALUE):
        if self.name + '-empty-marker' in self.request:
            self.updateWidgets(setErrors=False)

            # important: widget extract MUST return RAW values
            # just an extractData is WRONG here
            value, errors = self.extractRaw(setErrors=self.setErrors)

            if errors:
                # very-very-nasty: skip raising exceptions in extract
                # while we're updating -- that happens when the widget
                # is updated and update calls extract()
                if self._updating:
                    # don't rebind value, send back the original object
                    for name, widget_ in self.widgets.items():
                        if widget_.mode != interfaces.DISPLAY_MODE:
                            value[name] = widget_.value
                    return value
                raise MultipleErrors(errors)
            return value
        else:
            return default

    def render(self):
        """See z3c.form.interfaces.IWidget."""
        template = self.template
        if template is None:
            # one more discriminator than in widget.Widget
            template = zope.component.queryMultiAdapter(
                (self.context, self.request, self.form, self.field, self,
                 makeDummyObject(self.field.schema)),
                IPageTemplate, name=self.mode)
            if template is None:
                return super().render()
        return template(self)


# make dummy objects providing a given interface to support
# discriminating on field.schema

def makeDummyObject(iface):
    if iface is not None:
        @zope.interface.implementer(iface)
        class DummyObject:
            pass
    else:
        @zope.interface.implementer(zope.interface.Interface)
        class DummyObject:
            pass

    dummy = DummyObject()
    return dummy


# special template factory that takes the field.schema into account
# used by zcml.py

class ObjectWidgetTemplateFactory:
    """Widget template factory."""

    def __init__(self, filename, contentType='text/html',
                 context=None, request=None, view=None,
                 field=None, widget=None, schema=None):
        self.template = ViewPageTemplateFile(
            filename, content_type=contentType)
        zope.component.adapter(
            util.getSpecification(context),
            util.getSpecification(request),
            util.getSpecification(view),
            util.getSpecification(field),
            util.getSpecification(widget),
            util.getSpecification(schema))(self)
        zope.interface.implementer(IPageTemplate)(self)

    def __call__(self, context, request, view, field, widget, schema):
        return self.template


# default adapters

@zope.interface.implementer(interfaces.IObjectFactory)
class FactoryAdapter:
    """Most basic-default object factory adapter"""
    zope.component.adapts(
        zope.interface.Interface,  # context
        interfaces.IFormLayer,  # request
        zope.interface.Interface,
        # form -- but can become None easily (in tests)
        interfaces.IWidget)  # widget

    factory = None

    def __init__(self, context, request, form, widget):
        self.context = context
        self.request = request
        self.form = form
        self.widget = widget

    def __call__(self, value):
        # value is the extracted data from the form
        obj = self.factory()
        zope.event.notify(zope.lifecycleevent.ObjectCreatedEvent(obj))
        return obj


# XXX: Probably we should offer an register factory method which allows to
# use all discriminators e.g. context, request, form, widget as optional
# arguments. But can probably do that later in a ZCML directive
def registerFactoryAdapter(for_, klass):
    """register the basic FactoryAdapter for a given interface and class"""
    name = getIfName(for_)

    class temp(FactoryAdapter):
        factory = klass
    zope.component.provideAdapter(temp, name=name)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Common z3c.form test setups"""

import base64
import os
import pprint
import shutil
from doctest import register_optionflag

import lxml.doctestcompare
import lxml.html
import zope.browserresource
import zope.component
import zope.configuration.xmlconfig
import zope.interface
import zope.schema
from zope.browserpage.viewpagetemplatefile import BoundPageTemplate
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.pagetemplate.interfaces import IPageTemplate
from zope.publisher.browser import TestRequest
from zope.schema.fieldproperty import FieldProperty
from zope.security import checker
from zope.security.interfaces import IInteraction
from zope.security.interfaces import ISecurityPolicy

import z3c.form
from z3c.form import browser
from z3c.form import button
from z3c.form import contentprovider
from z3c.form import converter
from z3c.form import datamanager
from z3c.form import error
from z3c.form import field
from z3c.form import form
from z3c.form import interfaces
from z3c.form import outputchecker
from z3c.form import term
from z3c.form import tests
from z3c.form import validator
from z3c.form import widget
from z3c.form.browser import radio
from z3c.form.browser import select
from z3c.form.browser import text
from z3c.form.browser import textarea


# register lxml doctest option flags
lxml.doctestcompare.NOPARSE_MARKUP = register_optionflag('NOPARSE_MARKUP')

outputChecker = outputchecker.OutputChecker(patterns=())


class TestingFileUploadDataConverter(converter.FileUploadDataConverter):
    """A special file upload data converter that works with testing."""
    zope.component.adapts(
        zope.schema.interfaces.IBytes, interfaces.IFileWidget)

    def toFieldValue(self, value):
        if value is None or value == '':
            value = self.widget.request.get(self.widget.name + '.testing', '')
            encoding = self.widget.request.get(
                self.widget.name + '.encoding', 'plain')

            # allow for the case where the file contents are base64 encoded.
            if encoding == 'base64':
                value = base64.b64decode(value)
            self.widget.request.form[self.widget.name] = value

        return super().toFieldValue(value)


@zope.interface.implementer(interfaces.IFormLayer)
class TestRequest(TestRequest):
    def __init__(self, body_instream=None, environ=None, form=None,
                 skin=None, **kw):
        if form is not None:
            lists = {}
            for k in list(form.keys()):
                if isinstance(form[k], str):
                    # ensure unicode otherwise z3c.forms burps on str
                    form[k] = str(form[k])

                if k.endswith(':list'):
                    key = k.split(':', 1)[0]
                    lists.setdefault(key, []).append(form[k])
                    del form[k]

            form.update(lists)

        super().__init__(
            body_instream, environ, form, skin, **kw)


@zope.interface.implementer(IInteraction)
@zope.interface.provider(ISecurityPolicy)
class SimpleSecurityPolicy:
    """Allow all access."""

    loggedIn = False
    allowedPermissions = ()

    def __init__(self, loggedIn=False, allowedPermissions=()):
        self.loggedIn = loggedIn
        self.allowedPermissions = allowedPermissions + (checker.CheckerPublic,)

    def __call__(self, *participations):
        self.participations = []
        return self

    def checkPermission(self, permission, object):
        if self.loggedIn:
            if permission in self.allowedPermissions:
                return True
        return False


def getPath(filename):
    return os.path.join(os.path.dirname(browser.__file__), filename)


#############################
# classes required by ObjectWidget tests
#

class IMySubObject(zope.interface.Interface):
    foofield = zope.schema.Int(
        title="My foo field",
        default=1111,
        max=9999,
        required=True)
    barfield = zope.schema.Int(
        title="My dear bar",
        default=2222,
        required=False)


@zope.interface.implementer(IMySubObject)
class MySubObject:

    foofield = FieldProperty(IMySubObject['foofield'])
    barfield = FieldProperty(IMySubObject['barfield'])


class IMySecond(zope.interface.Interface):
    subfield = zope.schema.Object(
        title="Second-subobject",
        schema=IMySubObject)
    moofield = zope.schema.TextLine(title="Something")


@zope.interface.implementer(IMySecond)
class MySecond:

    subfield = FieldProperty(IMySecond['subfield'])
    moofield = FieldProperty(IMySecond['moofield'])


class IMyObject(zope.interface.Interface):
    subobject = zope.schema.Object(title='my object', schema=IMySubObject)
    name = zope.schema.TextLine(title='name')


@zope.interface.implementer(IMyObject)
class MyObject:

    def __init__(self, name='', subobject=None):
        self.subobject = subobject
        self.name = name


class IMyComplexObject(zope.interface.Interface):
    subobject = zope.schema.Object(title='my object', schema=IMySecond)
    name = zope.schema.TextLine(title='name')


class IMySubObjectMulti(zope.interface.Interface):
    foofield = zope.schema.Int(
        title="My foo field",
        default=None,  # default is None here!
        max=9999,
        required=True)
    barfield = zope.schema.Int(
        title="My dear bar",
        default=2222,
        required=False)


@zope.interface.implementer(IMySubObjectMulti)
class MySubObjectMulti:

    foofield = FieldProperty(IMySubObjectMulti['foofield'])
    barfield = FieldProperty(IMySubObjectMulti['barfield'])


class IMyMultiObject(zope.interface.Interface):
    listOfObject = zope.schema.List(
        title="My list field",
        value_type=zope.schema.Object(
            title='my object widget',
            schema=IMySubObjectMulti),
    )
    name = zope.schema.TextLine(title='name')


@zope.interface.implementer(IMyMultiObject)
class MyMultiObject:

    listOfObject = FieldProperty(IMyMultiObject['listOfObject'])
    name = FieldProperty(IMyMultiObject['name'])

    def __init__(self, name='', listOfObject=None):
        self.listOfObject = listOfObject
        self.name = name


def setUp(test):
    from zope.component.testing import setUp
    setUp()
    from zope.container.testing import setUp
    setUp()
    from zope.component import hooks
    hooks.setHooks()
    from zope.traversing.testing import setUp
    setUp()

    from zope.i18n.interfaces import IUserPreferredLanguages
    from zope.publisher.browser import BrowserLanguages
    from zope.publisher.interfaces.http import IHTTPRequest
    zope.component.provideAdapter(BrowserLanguages, [IHTTPRequest],
                                  IUserPreferredLanguages)

    from zope.site.folder import rootFolder
    site = rootFolder()
    from zope.component.interfaces import ISite
    from zope.site.site import LocalSiteManager
    if not ISite.providedBy(site):
        site.setSiteManager(LocalSiteManager(site))
    hooks.setSite(site)
    test.globs = {
        'root': site,
        'pprint': pprint.pprint}


def setUpZPT(suite):
    setUp(suite)


def setUpZ3CPT(suite):
    import z3c.pt
    import z3c.ptcompat
    setUp(suite)
    zope.configuration.xmlconfig.XMLConfig('configure.zcml', z3c.pt)()
    zope.configuration.xmlconfig.XMLConfig('configure.zcml', z3c.ptcompat)()


def setUpIntegration(test):
    setUp(test)
    zope.configuration.xmlconfig.XMLConfig('meta.zcml', zope.component)()
    zope.configuration.xmlconfig.XMLConfig('meta.zcml', zope.security)()
    zope.configuration.xmlconfig.XMLConfig('meta.zcml', zope.i18n)()
    zope.configuration.xmlconfig.XMLConfig('meta.zcml', zope.browserresource)()
    zope.configuration.xmlconfig.XMLConfig('meta.zcml', z3c.form)()
    zope.configuration.xmlconfig.XMLConfig('configure.zcml', z3c.form)()


def setupFormDefaults():
    # Validator adapters
    zope.component.provideAdapter(validator.SimpleFieldValidator)
    zope.component.provideAdapter(validator.InvariantsValidator)
    # Data manager adapter to get and set values to content
    zope.component.provideAdapter(datamanager.AttributeField)
    # Adapter to use form.fields to generate widgets
    zope.component.provideAdapter(field.FieldWidgets)
    # Adapter that uses form.fields to generate widgets
    # AND interlace content providers
    zope.component.provideAdapter(contentprovider.FieldWidgetsAndProviders)
    # Adapters to lookup the widget for a field
    # Text Field Widget
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.IBytesLine, interfaces.IFormLayer))
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.IASCIILine, interfaces.IFormLayer))
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.ITextLine, interfaces.IFormLayer))
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.IId, interfaces.IFormLayer))
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.IInt, interfaces.IFormLayer))
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.IFloat, interfaces.IFormLayer))
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.IDecimal, interfaces.IFormLayer))
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.IDate, interfaces.IFormLayer))
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.IDatetime, interfaces.IFormLayer))
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.ITime, interfaces.IFormLayer))
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.ITimedelta, interfaces.IFormLayer))
    zope.component.provideAdapter(
        text.TextFieldWidget,
        adapts=(zope.schema.interfaces.IURI, interfaces.IFormLayer))

    # Widget Layout
    zope.component.provideAdapter(
        widget.WidgetLayoutFactory(getPath('widget_layout.pt'), 'text/html'),
        (None, None, None, None, interfaces.IWidget),
        interfaces.IWidgetLayoutTemplate, name=interfaces.INPUT_MODE)
    zope.component.provideAdapter(
        widget.WidgetLayoutFactory(getPath('widget_layout.pt'), 'text/html'),
        (None, None, None, None, interfaces.IWidget),
        interfaces.IWidgetLayoutTemplate, name=interfaces.DISPLAY_MODE)
    zope.component.provideAdapter(
        widget.WidgetLayoutFactory(
            getPath('widget_layout_hidden.pt'), 'text/html'),
        (None, None, None, None, interfaces.IWidget),
        interfaces.IWidgetLayoutTemplate, name=interfaces.HIDDEN_MODE)

    # Text Field Widget
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(getPath('text_input.pt'), 'text/html'),
        (None, None, None, None, interfaces.ITextWidget),
        IPageTemplate, name=interfaces.INPUT_MODE)
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(getPath('text_display.pt'), 'text/html'),
        (None, None, None, None, interfaces.ITextWidget),
        IPageTemplate, name=interfaces.DISPLAY_MODE)
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(getPath('text_hidden.pt'), 'text/html'),
        (None, None, None, None, interfaces.ITextWidget),
        IPageTemplate, name=interfaces.HIDDEN_MODE)

    # Textarea Field Widget
    zope.component.provideAdapter(
        textarea.TextAreaFieldWidget,
        adapts=(zope.schema.interfaces.IASCII, interfaces.IFormLayer))
    zope.component.provideAdapter(
        textarea.TextAreaFieldWidget,
        adapts=(zope.schema.interfaces.IText, interfaces.IFormLayer))
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(
            getPath('textarea_input.pt'), 'text/html'),
        (None, None, None, None, interfaces.ITextAreaWidget),
        IPageTemplate, name=interfaces.INPUT_MODE)
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(
            getPath('textarea_display.pt'), 'text/html'),
        (None, None, None, None, interfaces.ITextAreaWidget),
        IPageTemplate, name=interfaces.DISPLAY_MODE)

    # Radio Field Widget
    zope.component.provideAdapter(radio.RadioFieldWidget)
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(getPath('radio_input.pt'), 'text/html'),
        (None, None, None, None, interfaces.IRadioWidget),
        IPageTemplate, name=interfaces.INPUT_MODE)
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(getPath('radio_display.pt'), 'text/html'),
        (None, None, None, None, interfaces.IRadioWidget),
        IPageTemplate, name=interfaces.DISPLAY_MODE)
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(getPath('radio_input_single.pt'),
                                     'text/html'),
        (None, None, None, None, interfaces.IRadioWidget),
        IPageTemplate, name='input_single')
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(getPath('radio_hidden_single.pt'),
                                     'text/html'),
        (None, None, None, None, interfaces.IRadioWidget),
        IPageTemplate, name='hidden_single')

    # Select Widget
    zope.component.provideAdapter(select.ChoiceWidgetDispatcher)
    zope.component.provideAdapter(select.SelectFieldWidget)
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(getPath('select_input.pt'), 'text/html'),
        (None, None, None, None, interfaces.ISelectWidget),
        IPageTemplate, name=interfaces.INPUT_MODE)
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(
            getPath('select_display.pt'), 'text/html'),
        (None, None, None, None, interfaces.ISelectWidget),
        IPageTemplate, name=interfaces.DISPLAY_MODE)
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(getPath('select_hidden.pt'), 'text/html'),
        (None, None, None, None, interfaces.ISelectWidget),
        IPageTemplate, name=interfaces.HIDDEN_MODE)

    # Checkbox Field Widget; register only templates
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(
            getPath('checkbox_input.pt'), 'text/html'),
        (None, None, None, None, interfaces.ICheckBoxWidget),
        IPageTemplate, name=interfaces.INPUT_MODE)
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(
            getPath('checkbox_display.pt'), 'text/html'),
        (None, None, None, None, interfaces.ICheckBoxWidget),
        IPageTemplate, name=interfaces.DISPLAY_MODE)
    # Submit Field Widget
    zope.component.provideAdapter(
        widget.WidgetTemplateFactory(getPath('submit_input.pt'), 'text/html'),
        (None, None, None, None, interfaces.ISubmitWidget),
        IPageTemplate, name=interfaces.INPUT_MODE)
    # selectwidget helper adapters
    zope.component.provideAdapter(select.CollectionSelectFieldWidget)
    zope.component.provideAdapter(select.CollectionChoiceSelectFieldWidget)
    # Adapter to  convert between field/internal and widget values
    zope.component.provideAdapter(converter.FieldDataConverter)
    zope.component.provideAdapter(converter.SequenceDataConverter)
    zope.component.provideAdapter(converter.CollectionSequenceDataConverter)
    zope.component.provideAdapter(converter.FieldWidgetDataConverter)
    # special data converter
    zope.component.provideAdapter(converter.IntegerDataConverter)
    zope.component.provideAdapter(converter.FloatDataConverter)
    zope.component.provideAdapter(converter.DecimalDataConverter)
    zope.component.provideAdapter(converter.DateDataConverter)
    zope.component.provideAdapter(converter.TimeDataConverter)
    zope.component.provideAdapter(converter.DatetimeDataConverter)
    zope.component.provideAdapter(converter.TimedeltaDataConverter)
    # Adapter for providing terms to radio list and other widgets
    zope.component.provideAdapter(term.BoolTerms)
    zope.component.provideAdapter(term.ChoiceTerms)
    zope.component.provideAdapter(term.ChoiceTermsVocabulary)
    zope.component.provideAdapter(term.ChoiceTermsSource)
    zope.component.provideAdapter(term.CollectionTerms)
    zope.component.provideAdapter(term.CollectionTermsVocabulary)
    zope.component.provideAdapter(term.CollectionTermsSource)
    # Adapter to create an action from a button
    zope.component.provideAdapter(
        button.ButtonAction, provides=interfaces.IButtonAction)
    # Adapter to use form.buttons to generate actions
    zope.component.provideAdapter(button.ButtonActions)
    # Adapter to use form.handlers to generate handle actions
    zope.component.provideAdapter(button.ButtonActionHandler)
    # Subscriber handling action execution error events
    zope.component.provideHandler(form.handleActionError)
    # Error View(s)
    zope.component.provideAdapter(error.ErrorViewSnippet)
    zope.component.provideAdapter(error.InvalidErrorViewSnippet)
    zope.component.provideAdapter(error.StandardErrorViewTemplate)


def tearDown(test):
    from zope.component import hooks
    from zope.testing import cleanup
    cleanup.cleanUp()
    hooks.resetHooks()
    hooks.setSite()


##########################
def render(view, xpath='.'):
    method = getattr(view, 'render', None)
    if method is None:
        method = view.__call__

    string = method()
    if string == "":
        return string

    try:
        root = lxml.etree.fromstring(string)
    except lxml.etree.XMLSyntaxError:
        root = lxml.html.fromstring(string)

    output = ""
    for node in root.xpath(
            xpath, namespaces={'xmlns': 'http://www.w3.org/1999/xhtml'}):
        s = lxml.etree.tounicode(node, pretty_print=True)
        s = s.replace(' xmlns="http://www.w3.org/1999/xhtml"', ' ')
        output += s

    if not output:
        raise ValueError("No elements matched by %s." % repr(xpath))

    # let's get rid of blank lines
    output = output.replace('\n\n', '\n')

    # self-closing tags are more readable with a space before the
    # end-of-tag marker
    output = output.replace('"/>', '" />')

    return output


def textOfWithOptionalTitle(node, addTitle=False, showTooltips=False):
    if isinstance(node, (list, tuple)):
        return '\n'.join(textOfWithOptionalTitle(child, addTitle, showTooltips)
                         for child in node)
    text = []
    if node is None:
        return None

    if node.tag == 'br':
        return '\n'
    if node.tag == 'input':
        if addTitle:
            title = node.get('name') or ''
            title += ' '
        else:
            title = ''
        if node.get('type') == 'radio':
            return title + ('(O)' if node.get('checked') else '( )')
        if node.get('type') == 'checkbox':
            return title + ('[x]' if node.get('checked') else '[ ]')
        if node.get('type') == 'hidden':
            return ''
        else:
            return '{}[{}]'.format(title, node.get('value') or '')
    if node.tag == 'textarea':
        if addTitle:
            title = node.get('name') or ''
            title += ' '
            text.append(title)
    if node.tag == 'select':
        if addTitle:
            title = node.get('name') or ''
            title += ' '
        else:
            title = ''
        option = node.find('option[@selected]')
        return '{}[{}]'.format(
            title,
            option.text if option is not None else '[    ]')
    if node.tag == 'li':
        text.append('*')
    if node.tag == 'script':
        return

    if node.text and node.text.strip():
        text.append(node.text.strip())

    for n, child in enumerate(node):
        s = textOfWithOptionalTitle(child, addTitle, showTooltips)
        if s:
            text.append(s)
        if child.tail and child.tail.strip():
            text.append(child.tail)
    text = ' '.join(text).strip()
    # 'foo<br>bar' ends up as 'foo \nbar' due to the algorithm used above
    text = text.replace(' \n', '\n').replace('\n ', '\n').replace('\n\n', '\n')
    if '\xA0' in text:
        # don't just .replace, that'll sprinkle my tests with u''
        text = text.replace('\xA0', ' ')  # nbsp -> space
    if node.tag == 'li':
        text += '\n'
    if node.tag == 'div':
        text += '\n'
    return text


def textOf(node):
    """Return the contents of an HTML node as text.

    Useful for functional tests, e.g. ::

        print map(textOf, browser.etree.xpath('.//td'))

    """
    return textOfWithOptionalTitle(node, False)


def plainText(content, xpath=None):
    root = lxml.html.fromstring(content)
    if xpath is not None:
        nodes = root.xpath(xpath)
        joinon = '\n'
    else:
        nodes = root
        joinon = ''
    text = joinon.join(map(textOf, nodes))
    lines = [line.strip() for line in text.splitlines()]
    text = '\n'.join(lines)
    return text


def getSubmitValues(content):
    root = lxml.html.fromstring(content)
    form = root.forms[0]
    values = dict(form.form_values())
    return values


def addTemplate(form, fname):
    form.template = BoundPageTemplate(
        ViewPageTemplateFile(
            fname, os.path.dirname(tests.__file__)), form)


def saveHtml(content, fname):
    path = os.path.join(os.getcwd(), 'var', 'htmls')
    if not os.path.exists(path):
        os.makedirs(path)
    fullfname = os.path.join(path, fname)
    with open(fullfname, 'wb') as fout:
        fout.write(content.encode('utf8'))

    fullfname = os.path.join(path, 'integration.css')
    cssfname = os.path.join(os.path.dirname(tests.__file__),
                            'integration.css')
    shutil.copy(cssfname, fullfname)


##########################################
# integration test interfaces and classes

class IntegrationBase:
    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)

    def __repr__(self):
        items = sorted(self.__dict__.items())
        return ("<" + self.__class__.__name__ + "\n  "
                + "\n  ".join([f"{key}: {pprint.pformat(value)}"
                               for key, value in items]) + ">")


class IObjectWidgetSingleSubIntegration(zope.interface.Interface):
    singleInt = zope.schema.Int(
        title='Int label')
    singleBool = zope.schema.Bool(
        title='Bool label',
        required=True)
    singleChoice = zope.schema.Choice(
        title='Choice label',
        values=('one', 'two', 'three'))
    singleChoiceOpt = zope.schema.Choice(
        title='ChoiceOpt label',
        values=('four', 'five', 'six'),
        required=False)
    singleTextLine = zope.schema.TextLine(
        title='TextLine label')
    singleDate = zope.schema.Date(
        title='Date label')
    singleReadOnly = zope.schema.TextLine(
        title='ReadOnly label',
        readonly=True)


@zope.interface.implementer(IObjectWidgetSingleSubIntegration)
class ObjectWidgetSingleSubIntegration(IntegrationBase):

    singleInt = FieldProperty(IObjectWidgetSingleSubIntegration['singleInt'])
    singleBool = FieldProperty(IObjectWidgetSingleSubIntegration['singleBool'])
    singleChoice = FieldProperty(
        IObjectWidgetSingleSubIntegration['singleChoice'])
    singleChoiceOpt = FieldProperty(
        IObjectWidgetSingleSubIntegration['singleChoiceOpt'])
    singleTextLine = FieldProperty(
        IObjectWidgetSingleSubIntegration['singleTextLine'])
    singleDate = FieldProperty(IObjectWidgetSingleSubIntegration['singleDate'])
    singleReadOnly = FieldProperty(
        IObjectWidgetSingleSubIntegration['singleReadOnly'])


class IObjectWidgetSingleIntegration(zope.interface.Interface):
    subobj = zope.schema.Object(
        title='Object label',
        schema=IObjectWidgetSingleSubIntegration
    )


@zope.interface.implementer(IObjectWidgetSingleIntegration)
class ObjectWidgetSingleIntegration:

    subobj = FieldProperty(IObjectWidgetSingleIntegration['subobj'])


class IObjectWidgetMultiSubIntegration(zope.interface.Interface):
    multiInt = zope.schema.Int(
        title='Int label')
    multiBool = zope.schema.Bool(
        title='Bool label',
        required=True)
    multiChoice = zope.schema.Choice(
        title='Choice label',
        values=('one', 'two', 'three'))
    multiChoiceOpt = zope.schema.Choice(
        title='ChoiceOpt label',
        values=('four', 'five', 'six'),
        required=False)
    multiTextLine = zope.schema.TextLine(
        title='TextLine label')
    multiDate = zope.schema.Date(
        title='Date label')


@zope.interface.implementer(IObjectWidgetMultiSubIntegration)
class ObjectWidgetMultiSubIntegration(IntegrationBase):

    multiInt = FieldProperty(IObjectWidgetMultiSubIntegration['multiInt'])
    multiBool = FieldProperty(IObjectWidgetMultiSubIntegration['multiBool'])
    multiChoice = FieldProperty(
        IObjectWidgetMultiSubIntegration['multiChoice'])
    multiChoiceOpt = FieldProperty(
        IObjectWidgetMultiSubIntegration['multiChoiceOpt'])
    multiTextLine = FieldProperty(
        IObjectWidgetMultiSubIntegration['multiTextLine'])
    multiDate = FieldProperty(IObjectWidgetMultiSubIntegration['multiDate'])


class IObjectWidgetMultiIntegration(zope.interface.Interface):
    subobj = zope.schema.Object(
        title='Object label',
        schema=IObjectWidgetMultiSubIntegration
    )


@zope.interface.implementer(IObjectWidgetMultiIntegration)
class ObjectWidgetMultiIntegration:

    subobj = FieldProperty(IObjectWidgetMultiIntegration['subobj'])


class IMultiWidgetListIntegration(zope.interface.Interface):
    listOfInt = zope.schema.List(
        title="ListOfInt label",
        value_type=zope.schema.Int(
            title='Int label'),
    )
    listOfBool = zope.schema.List(
        title="ListOfBool label",
        value_type=zope.schema.Bool(
            title='Bool label',
            required=True),
    )
    listOfChoice = zope.schema.List(
        title="ListOfChoice label",
        value_type=zope.schema.Choice(
            title='Choice label',
            values=('one', 'two', 'three')
        ),
    )
    listOfTextLine = zope.schema.List(
        title="ListOfTextLine label",
        value_type=zope.schema.TextLine(
            title='TextLine label'),
    )
    listOfDate = zope.schema.List(
        title="ListOfDate label",
        value_type=zope.schema.Date(
            title='Date label'),
    )
    listOfObject = zope.schema.List(
        title="ListOfObject label",
        value_type=zope.schema.Object(
            title='Object label',
            schema=IObjectWidgetMultiSubIntegration),
    )


@zope.interface.implementer(IMultiWidgetListIntegration)
class MultiWidgetListIntegration(IntegrationBase):

    listOfInt = FieldProperty(IMultiWidgetListIntegration['listOfInt'])
    listOfBool = FieldProperty(IMultiWidgetListIntegration['listOfBool'])
    listOfChoice = FieldProperty(IMultiWidgetListIntegration['listOfChoice'])
    listOfTextLine = FieldProperty(
        IMultiWidgetListIntegration['listOfTextLine'])
    listOfDate = FieldProperty(IMultiWidgetListIntegration['listOfDate'])
    listOfObject = FieldProperty(IMultiWidgetListIntegration['listOfObject'])


class IMultiWidgetDictIntegration(zope.interface.Interface):
    dictOfInt = zope.schema.Dict(
        title="DictOfInt label",
        key_type=zope.schema.Int(
            title='Int key'),
        value_type=zope.schema.Int(
            title='Int label'),
    )
    dictOfBool = zope.schema.Dict(
        title="DictOfBool label",
        key_type=zope.schema.Bool(
            title='Bool key',
            required=True),
        value_type=zope.schema.Bool(
            title='Bool label',
            required=True),
    )
    dictOfChoice = zope.schema.Dict(
        title="DictOfChoice label",
        key_type=zope.schema.Choice(
            title='Choice key',
            values=('key1', 'key2', 'key3')
        ),
        value_type=zope.schema.Choice(
            title='Choice label',
            values=('one', 'two', 'three')
        ),
    )
    dictOfTextLine = zope.schema.Dict(
        title="DictOfTextLine label",
        key_type=zope.schema.TextLine(
            title='TextLine key'),
        value_type=zope.schema.TextLine(
            title='TextLine label'),
    )
    dictOfDate = zope.schema.Dict(
        title="DictOfDate label",
        key_type=zope.schema.Date(
            title='Date key'),
        value_type=zope.schema.Date(
            title='Date label'),
    )
    dictOfObject = zope.schema.Dict(
        title="DictOfObject label",
        key_type=zope.schema.TextLine(
            title='Object key'),
        value_type=zope.schema.Object(
            title='Object label',
            schema=IObjectWidgetMultiSubIntegration),
    )


@zope.interface.implementer(IMultiWidgetDictIntegration)
class MultiWidgetDictIntegration(IntegrationBase):

    dictOfInt = FieldProperty(IMultiWidgetDictIntegration['dictOfInt'])
    dictOfBool = FieldProperty(IMultiWidgetDictIntegration['dictOfBool'])
    dictOfChoice = FieldProperty(IMultiWidgetDictIntegration['dictOfChoice'])
    dictOfTextLine = FieldProperty(
        IMultiWidgetDictIntegration['dictOfTextLine'])
    dictOfDate = FieldProperty(IMultiWidgetDictIntegration['dictOfDate'])
    dictOfObject = FieldProperty(IMultiWidgetDictIntegration['dictOfObject'])



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Unittests for bug fixes."""

import unittest


class TestApplyChangesDictDatamanager(unittest.TestCase):
    # z3c.form.form.applyChanges could not write a value into an empty
    # content dict it got an AttributeError while accessing
    # datamanger.get(). This test makes sure that a dictionary content
    # does not need to be initialized with all keys which might be
    # written on it later on. (This was the behavior before
    # datamanger.DictionaryField.get() raised an AttributeError on not
    # existing keys.

    def setUp(self):
        import zope.component
        import zope.interface

        import z3c.form.datamanager

        zope.component.provideAdapter(
            z3c.form.datamanager.DictionaryField,
            (dict, zope.interface.Interface))

    def tearDown(self):
        import zope.component.globalregistry

        import z3c.form.datamanager

        zope.component.globalregistry.getGlobalSiteManager().unregisterAdapter(
            z3c.form.datamanager.DictionaryField,
            (dict, zope.interface.Interface))

    def test_applyChanges(self):
        import zope.interface
        import zope.schema

        import z3c.form.field
        import z3c.form.form

        class TestInterface(zope.interface.Interface):
            text = zope.schema.TextLine(title='text')

        class TestForm(z3c.form.form.BaseForm):
            fields = z3c.form.field.Fields(TestInterface)

        # content is an empty dict, the `text` key does not yet exist
        content = dict()
        form = TestForm(content, request=None)
        data = dict(text='a')
        changes = z3c.form.form.applyChanges(form, content, data)
        self.assertEqual({TestInterface: ['text']}, changes)
        self.assertEqual({'text': 'a'}, content)


class Mock:
    pass


class MockNumberFormatter:
    def format(self, value):
        if value is None:
            # execution should never get here
            raise ValueError('Cannot format None')
        return str(value)


class MockLocale:
    def getFormatter(self, category):
        return MockNumberFormatter()


class OurNone:
    def __eq__(self, other):
        return isinstance(other, (type(None), OurNone))


OUR_NONE = OurNone()
# OUR_NONE == None
# but
# OUR_NONE is not None


class ConverterFixTests(unittest.TestCase):
    # most of the time `==` and `is` works the same
    # unless you have some custom or mutable values

    def test_BaseDataConverter_toWidgetValue(self):
        from z3c.form.converter import BaseDataConverter

        field = Mock()
        field.missing_value = None
        bdc = BaseDataConverter(field, None)
        self.assertEqual(bdc.toWidgetValue(''), '')
        self.assertEqual(bdc.toWidgetValue(None), '')
        self.assertEqual(bdc.toWidgetValue([]), '[]')

        field.missing_value = []
        self.assertEqual(bdc.toWidgetValue(''), '')
        self.assertEqual(bdc.toWidgetValue(None), 'None')
        self.assertEqual(bdc.toWidgetValue([]), '')

    def test_NumberDataConverter_toWidgetValue(self):
        from z3c.form.converter import NumberDataConverter

        field = Mock()
        field.missing_value = None
        widget = Mock()
        widget.request = Mock()
        widget.request.locale = Mock()
        widget.request.locale.numbers = MockLocale()
        ndc = NumberDataConverter(field, widget)
        self.assertEqual(ndc.toWidgetValue(''), '')
        self.assertEqual(ndc.toWidgetValue(None), '')
        # here is the real deal, OUR_NONE should be considered as None
        self.assertEqual(ndc.toWidgetValue(OUR_NONE), '')
        self.assertEqual(ndc.toWidgetValue([]), '[]')

        field.missing_value = OUR_NONE
        self.assertEqual(ndc.toWidgetValue(''), '')
        self.assertEqual(ndc.toWidgetValue(None), '')
        self.assertEqual(ndc.toWidgetValue(OUR_NONE), '')
        self.assertEqual(ndc.toWidgetValue([]), '[]')


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



import zope.interface

from z3c.form.interfaces import IDataExtractedEvent


@zope.interface.implementer(IDataExtractedEvent)
class DataExtractedEvent:

    def __init__(self, data, errors, form):
        self.data = data
        self.errors = errors
        self.form = form



from setuptools import find_packages
from setuptools import setup


version = '0.1'

setup(name='benchmark',
      version=version,
      description="Benchmark-suite for z3c.form.",
      long_description="""\
      """,
      keywords='',
      author="Malthe Borch and the Zope Community",
      author_email="zope-dev@zope.org",
      url='',
      license='',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'zope.app.pagetemplate',
          'zope.schema',
          'z3c.form',
          'z3c.pt',
          'z3c.ptcompat',
      ],
      )
